"""Patito, a data-modelling library built on top of polars and pydantic."""
import pkg_resources
from pkg_resources import DistributionNotFound
from polars import Expr, Series, col

from patito import exceptions
from patito.exceptions import ValidationError
from patito.polars import DataFrame
from patito.pydantic import Field, Model

_DUCKDB_AVAILABLE = False
__all__ = [
    "DataFrame",
    "Expr",
    "Field",
    "Model",
    "Series",
    "ValidationError",
    "col",
    "exceptions",
    "_DUCKDB_AVAILABLE",
]

try:
    pkg_resources.require(["duckdb>=0.3.2"])
    from patito.duckdb import Database, Relation, RelationSource

    _DUCKDB_AVAILABLE = True
    __all__ += [
        "Database",
        "Relation",
        "RelationSource",
    ]
except DistributionNotFound:  # pragma: no cover
    pass


try:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import PackageNotFoundError, version  # type: ignore

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



"""Logic related to wrapping logic around the pydantic library."""
from __future__ import annotations

import itertools
from collections.abc import Iterable
from datetime import date, datetime
from typing import Any, ClassVar, Dict, List, Optional, Set, Type, TypeVar, Union

import polars as pl
from pydantic import BaseConfig, BaseModel, Field  # noqa: F401
from pydantic.main import ModelMetaclass as PydanticModelMetaclass

from patito.polars import DataFrame
from patito.validators import validate

try:
    import pandas as pd

    _PANDAS_AVAILABLE = True
except ImportError:
    _PANDAS_AVAILABLE = False

# The generic type of a single row in given Relation.
# Should be a typed subclass of Model.
ModelType = TypeVar("ModelType", bound="Model")

# A mapping from pydantic types to the equivalent type used in DuckDB
PYDANTIC_TO_DUCKDB_TYPES = {
    "integer": "BIGINT",
    "string": "VARCHAR",
    "number": "DOUBLE",
    "boolean": "BOOLEAN",
}

# A mapping from pydantic types to equivalent dtypes used in polars
PYDANTIC_TO_POLARS_TYPES = {
    "integer": pl.Int64,
    "string": pl.Utf8,
    "number": pl.Float64,
    "boolean": pl.Boolean,
}


class ModelMetaclass(PydanticModelMetaclass):
    """
    Metclass used by patito.Model.

    Responsible for setting any relevant model-dependent class properties.
    """

    def __init__(cls, name: str, bases: tuple, clsdict: dict) -> None:
        """
        Construct new patito model.

        Args:
            name: Name of model class.
            bases: Tuple of superclasses.
            clsdict: Dictionary containing class properties.
        """
        super().__init__(name, bases, clsdict)  # type: ignore
        # Add a custom subclass of patito.DataFrame to the model class,
        # where .set_model() has been implicitly set.
        cls.DataFrame = DataFrame._construct_dataframe_model_class(
            model=cls,  # type: ignore
        )

    # --- Class properties ---
    # These properties will only be available on Model *classes*, not instantiated
    # objects This is backwards compatible to python versions before python 3.9,
    # unlike a combination of @classmethod and @property.
    @property
    def columns(cls: Type[ModelType]) -> List[str]:  # type: ignore
        """
        Return the name of the specified column fields the DataFrame.

        Returns:
            List of column names.
        """
        return list(cls.schema()["properties"].keys())

    @property
    def dtypes(  # type: ignore
        cls: Type[ModelType],
    ) -> dict[str, Type[pl.DataType]]:
        """
        Return the dtypes of the dataframe.

        Unless Field(dtype=...) is specified, the highest signed column dtype
        is chosen for integer and float columns.

        Returns:
            A dictionary mapping string column names to polars dtype classes.
        """
        return {
            column: valid_dtypes[0] for column, valid_dtypes in cls.valid_dtypes.items()
        }

    @property
    def valid_dtypes(  # type: ignore  # noqa: C901
        cls: Type[ModelType],
    ) -> dict[str, List[Type[pl.DataType]]]:
        """
        Return valid polars dtypes as a column name -> dtypes mapping.

        The first item of each list is the default dtype chosen by Patito.

        Returns:
            A dictionary mapping each column string name to a list of valid
            dtypes.

        Raises:
            NotImplementedError: If one or more model fields are annotated with types
                not compatible with polars.
        """
        schema = cls.schema()
        properties = schema["properties"]

        valid_dtypes = {}
        for column, props in properties.items():
            if "dtype" in props:
                valid_dtypes[column] = [
                    props["dtype"],
                ]
            elif "enum" in props:
                if props["type"] != "string":  # pragma: no cover
                    raise NotImplementedError
                valid_dtypes[column] = [pl.Categorical, pl.Utf8]
            elif "type" not in props:
                raise NotImplementedError(
                    f"No valid dtype mapping found for column '{column}'."
                )
            elif props["type"] == "integer":
                valid_dtypes[column] = [
                    pl.Int64,
                    pl.Int32,
                    pl.Int16,
                    pl.Int8,
                    pl.UInt64,
                    pl.UInt32,
                    pl.UInt16,
                    pl.UInt8,
                ]
            elif props["type"] == "number":
                if props.get("format") == "time-delta":
                    valid_dtypes[column] = [
                        pl.Duration,
                    ]  # pyright: reportPrivateImportUsage=false
                else:
                    valid_dtypes[column] = [pl.Float64, pl.Float32]
            elif props["type"] == "boolean":
                valid_dtypes[column] = [
                    pl.Boolean,
                ]
            elif props["type"] == "string":
                string_format = props.get("format")
                if string_format is None:
                    valid_dtypes[column] = [
                        pl.Utf8,
                    ]
                elif string_format == "date":
                    valid_dtypes[column] = [
                        pl.Date,
                    ]
                # TODO: Find out why this branch is not being hit
                elif string_format == "date-time":  # pragma: no cover
                    valid_dtypes[column] = [
                        pl.Datetime,
                    ]
            elif props["type"] == "null":
                valid_dtypes[column] = [
                    pl.Null,
                ]
            else:  # pragma: no cover
                raise NotImplementedError(
                    f"No valid dtype mapping found for column '{column}'"
                )

        return valid_dtypes

    @property
    def defaults(  # type: ignore
        cls: Type[ModelType],
    ) -> dict[str, Any]:
        """
        Return default field values specified on the model.

        Returns:
            Dictionary containing fields with their respective default values.
        """
        return {
            field_name: props["default"]
            for field_name, props in cls.schema()["properties"].items()
            if "default" in props
        }

    @property
    def non_nullable_columns(  # type: ignore
        cls: Type[ModelType],  # pyright: reportGeneralTypeIssues=false
    ) -> set[str]:
        """
        Return names of those columns that are non-nullable in the schema.

        Returns:
            Set of column name strings.
        """
        return set(cls.schema()["required"])

    @property
    def nullable_columns(  # type: ignore
        cls: Type[ModelType],  # pyright: reportGeneralTypeIssues=false
    ) -> set[str]:
        """
        Return names of those columns that are nullable in the schema.

        Returns:
            Set of column name strings.
        """
        return set(cls.columns) - cls.non_nullable_columns

    @property
    def unique_columns(  # type: ignore
        cls: Type[ModelType],
    ) -> set[str]:
        """
        Return columns with uniqueness constraint.

        Returns:
            Set of column name strings.
        """
        props = cls.schema()["properties"]
        return {column for column in cls.columns if props[column].get("unique", False)}

    @property
    def sql_types(  # type: ignore
        cls: Type[ModelType],
    ) -> dict[str, str]:
        """
        Return SQL types as a column name -> sql type dict mapping.

        Returns:
            Dictionary with column name keys and SQL type identifier strings.
        """
        schema = cls.schema()
        props = schema["properties"]
        return {
            column: PYDANTIC_TO_DUCKDB_TYPES[props[column]["type"]]
            for column in cls.columns
        }


class Model(BaseModel, metaclass=ModelMetaclass):
    """Custom pydantic class for representing table schema and constructing rows."""

    # -- Class properties set by model metaclass --
    # This weird combination of a MetaClass + type annotation
    # in order to make the following work simultaneously:
    #     1. Make these dynamically constructed properties of the class.
    #     2. Have the correct type information for type checkers.
    #     3. Allow sphinx-autodoc to construct correct documentation.
    #     4. Be compatible with python 3.7.
    # Once we drop support for python 3.7, we can replace all of this with just a simple
    # combination of @property and @classmethod.
    columns: ClassVar[List[str]]

    unique_columns: ClassVar[Set[str]]
    non_nullable_columns: ClassVar[Set[str]]
    nullable_columns: ClassVar[Set[str]]

    dtypes: ClassVar[Dict[str, Type[pl.DataType]]]
    sql_types: ClassVar[Dict[str, str]]
    valid_dtypes: ClassVar[Dict[str, List[Type[pl.DataType]]]]

    defaults: ClassVar[Dict[str, Any]]

    @classmethod
    @property
    def DataFrame(cls: Type[ModelType]) -> Type[DataFrame[ModelType]]:  # type: ignore
        """Return DataFrame class where DataFrame.set_model() is set to self."""

    @classmethod
    def from_row(
        cls: Type[ModelType],
        row: Union["pd.DataFrame", pl.DataFrame],
        validate: bool = True,
    ) -> ModelType:
        """
        Represent a single data frame row as a patito model.

        Args:
            row: A dataframe, either polars and pandas, consisting of a single row.
            validate: If False, skip pydantic validation of the given row data.

        Returns:
            A patito model representing the given row data.

        Raises:
            TypeError: If the given type is neither a pandas or polars DataFrame.
        """
        if isinstance(row, pl.DataFrame):
            dataframe = row
        elif _PANDAS_AVAILABLE and isinstance(row, pd.DataFrame):
            dataframe = pl.DataFrame._from_pandas(row)
        elif _PANDAS_AVAILABLE and isinstance(row, pd.Series):
            return cls(**dict(row.iteritems()))
        else:
            raise TypeError(f"{cls.__name__}.from_row not implemented for {type(row)}.")
        return cls.from_polars(dataframe=dataframe, validate=validate)

    @classmethod
    def from_polars(
        cls: Type[ModelType],
        dataframe: pl.DataFrame,
        validate: bool = True,
    ) -> ModelType:
        """
        Construct model from a single polars row.

        Args:
            dataframe: A polars dataframe consisting of one single row.
            validate: If True, run the pydantic validators. If False, pydantic will
                not cast any types in the resulting object.

        Returns:
            A pydantic model object representing the given polars row.

        Raises:
            TypeError: If the provided `dataframe` argument is not of type
                polars.DataFrame.
            ValueError: If the given `dataframe` argument does not consist of exactly
                one row.
        """
        if not isinstance(dataframe, pl.DataFrame):
            raise TypeError(
                f"{cls.__name__}.from_polars() must be invoked with polars.DataFrame, "
                f"not {type(dataframe)}!"
            )
        elif len(dataframe) != 1:
            raise ValueError(
                f"{cls.__name__}.from_polars() can only be invoked with exactly "
                f"1 row, while {len(dataframe)} rows were provided."
            )

        # We have been provided with a single polars.DataFrame row
        # Convert to the equivalent keyword invocation of the pydantic model
        if validate:
            return cls(**dataframe.to_dicts()[0])
        else:
            return cls.construct(**dataframe.to_dicts()[0])

    @classmethod
    def validate(  # type: ignore
        cls,
        dataframe: Union["pd.DataFrame", pl.DataFrame],
    ) -> None:
        """
        Validate the given dataframe.

        Args:
            dataframe: Polars DataFrame to be validated.

        Raises:
            patito.exceptions.ValidationError:  # noqa: DAR402
                If the given dataframe does not match the given schema.
        """
        validate(dataframe=dataframe, schema=cls)

    @classmethod
    def example_value(  # noqa: C901
        cls,
        field: str,
    ) -> Union[date, datetime, float, int, str, None]:
        """
        Return an example value for the given field name defined on the model.

        Args:
            field: Field name identifier.

        Returns:
            A single value which is consistent with the given field definition.

        Raises:
            NotImplementedError: If the given field has no example generator.
        """
        schema = cls.schema()
        field_data = schema["properties"]
        non_nullable = schema["required"]
        properties = field_data[field]
        field_type = properties["type"]
        if "const" in properties:
            # The default value is the only valid value, provided as const
            return properties["const"]

        elif "default" in properties:
            # A default value has been specified in the model field definiton
            return properties["default"]

        elif "enum" in properties:
            return properties["enum"][0]

        elif field not in non_nullable:
            return None

        elif field_type in {"integer", "number"}:
            # For integer and float types we must check if there are imposed bounds
            lower = properties.get("minimum") or properties.get("exclusiveMinimum")
            upper = properties.get("maximum") or properties.get("exclusiveMaximum")

            # If the dtype is an unsigned integer type, we must return a positive value
            if "dtype" in properties:
                dtype = properties["dtype"]
                if dtype in (pl.UInt8, pl.UInt16, pl.UInt32, pl.UInt64):
                    lower = 0 if lower is None else max(lower, 0)

            # First we check the simple case, no upper or lower bound
            if lower is None and upper is None:
                if field_type == "number":
                    return -0.5
                else:
                    return -1

            # If we have a lower and upper bound, we return something in the middle
            elif lower is not None and upper is not None:
                if field_type == "number":
                    return (lower + upper) / 2
                else:
                    return (lower + upper) // 2

            # What remains is a single-sided bound, which we will return a value on the
            # "right side" of.
            number = float if field_type == "number" else int
            if lower is not None:
                return number(lower + 1)
            else:
                return number(upper - 1)

        elif field_type == "string":
            if "pattern" in properties:
                raise NotImplementedError(
                    "Example data generation has not been implemented for regex "
                    "patterns. You must valid data for such columns explicitly!"
                )
            elif "format" in properties and properties["format"] == "date":
                return date(year=1970, month=1, day=1)
            elif "format" in properties and properties["format"] == "date-time":
                return datetime(year=1970, month=1, day=1)
            elif "minLength" in properties:
                return "a" * properties["minLength"]
            elif "maxLength" in properties:
                return "a" * min(properties["maxLength"], 1)
            else:
                return "dummy_string"

        elif field_type == "boolean":
            return False

        else:  # pragma: no cover
            raise NotImplementedError

    @classmethod
    def example(
        cls: Type[ModelType],
        **kwargs: Any,  # noqa: ANN401
    ) -> ModelType:
        """
        Produce model with dummy data for all unspecified fields.

        The type annotation of unspecified field is used to fill in type-correct
        dummy data, e.g. -1 for int, "dummy_string" for str, and so on...
        The first item of Literal annotatations are used for dummy values.

        Args:
            **kwargs: Provide explicit values for any fields which should not be filled
                with dummy data.

        Returns:
            A pydantic model object filled with dummy data for all unspecified model
                fields.

        Raises:
            TypeError: If one or more of the provided keyword arguments do not match any
                fields on the model.
        """
        # Non-iterable values besides strings must be repeated
        wrong_columns = set(kwargs.keys()) - set(cls.columns)
        if wrong_columns:
            raise TypeError(f"{cls.__name__} does not contain fields {wrong_columns}!")

        schema = cls.schema()
        properties = schema["properties"]
        new_kwargs = {}
        for field_name in properties.keys():
            if field_name in kwargs:
                # The value has been explicitly specified
                new_kwargs[field_name] = kwargs[field_name]
            else:
                new_kwargs[field_name] = cls.example_value(field=field_name)
        return cls(**new_kwargs)

    @classmethod
    def pandas_examples(
        cls: Type[ModelType],
        data: Union[dict, Iterable],
        columns: Optional[Iterable[str]] = None,
    ) -> "pd.DataFrame":
        """
        Generate dataframe with dummy data for all unspecified columns.

        Offers the same API as the pandas.DataFrame constructor.
        Non-iterable values, besides strings, are repeated until they become as long as
        the iterable arguments.

        Args:
            data: Data to populate the dummy dataframe with. If
                not a dict, column names must also be provided.
            columns: Ignored if data is a dict. If
                data is an iterable, it will be used as the column names in the
                resulting dataframe. Defaults to None.

        Returns:
            A pandas DataFrame filled with dummy example data.

        Raises:
            ImportError: If pandas has not been installed. You should install
                patito[pandas] in order to integrate patito with pandas.
            TypeError: If column names have not been specified in the input data.
        """
        if not _PANDAS_AVAILABLE:
            # Re-trigger the import error, but this time don't catch it
            raise ImportError("No module named 'pandas'")

        if not isinstance(data, dict):
            if columns is None:
                raise TypeError(
                    f"{cls.__name__}.pandas_examples() must "
                    "be provided with column names!"
                )
            kwargs = dict(zip(columns, zip(*data)))
        else:
            kwargs = data

        kwargs = {
            key: (
                value
                if isinstance(value, Iterable) and not isinstance(value, str)
                else itertools.cycle([value])
            )
            for key, value in kwargs.items()
        }
        dummies = []
        for values in zip(*kwargs.values()):
            dummies.append(cls.example(**dict(zip(kwargs.keys(), values))))
        return pd.DataFrame([dummy.dict() for dummy in dummies])

    @classmethod
    def examples(
        cls: Type[ModelType],
        data: Optional[Union[dict, Iterable]] = None,
        columns: Optional[Iterable[str]] = None,
    ) -> pl.DataFrame:
        """
        Generate polars dataframe with dummy data for all unspecified columns.

        This constructor accepts the same data format as polars.DataFrame.

        Args:
            data: Data to populate the dummy dataframe with. If given as an iterable of
                values then column names must also be provided. If not provided at all,
                an empty dataframe with the correct column dtypes will be generated
                instead.
            columns: Ignored if data is a dict. If data is an iterable, it will be used
                as the column names in the resulting dataframe. Defaults to None.

        Returns:
            A polars dataframe where all unspecified columns have been filled with dummy
                data which should pass model validation.

        Raises:
            TypeError: If one or more of the model fields are not mappable to polars
                column dtype equivalents.
        """
        if data is None:
            # We should create an empty dataframe, but with the correct dtypes
            kwargs = {}
        elif not isinstance(data, dict):
            if columns is None:
                raise TypeError(
                    f"{cls.__name__}.examples() must be provided with column names!"
                )
            kwargs = dict(zip(columns, zip(*data)))
        else:
            kwargs = data

        wrong_columns = set(kwargs.keys()) - set(cls.columns)
        if wrong_columns:
            raise TypeError(f"{cls.__name__} does not contain fields {wrong_columns}!")

        series: List[Union[pl.Series, pl.Expr]] = []
        unique_series = []
        for column_name, dtype in cls.dtypes.items():
            if column_name not in kwargs:
                if column_name in cls.unique_columns:
                    unique_series.append(
                        pl.first().cumcount().cast(dtype).alias(column_name)
                    )
                else:
                    example_value = cls.example_value(field=column_name)
                    series.append(pl.lit(example_value, dtype=dtype).alias(column_name))
                continue

            value = kwargs.get(column_name)
            if isinstance(value, Iterable) and not isinstance(value, str):
                # We make sure that at least one series is inserted first in the list,
                # otherwise polars will not be able to handle the shape mismatch between
                # series and literate values.
                series.insert(0, pl.Series(name=column_name, values=value, dtype=dtype))
            else:
                series.append(pl.lit(value, dtype=dtype).alias(column_name))

        return (
            pl.DataFrame()
            .with_columns(series)  # type: ignore
            .with_columns(unique_series)
        )

    class Config(BaseConfig):
        """Configuration for Pydantic BaseModel behaviour."""

        # Make fields immutable and model hashable
        frozen = True



"""Module containing all custom exceptions raised by patito."""

import pydantic


class ValidationError(pydantic.ValidationError):
    """Exception raised when dataframe does not match schema."""


class ErrorWrapper(pydantic.error_wrappers.ErrorWrapper):
    """Wrapper for specific column validation error."""


class WrongColumnsError(TypeError):
    """Validation exception for column name mismatches."""


class MissingColumnsError(WrongColumnsError):
    """Exception for when a dataframe is missing one or more columns."""


class SuperflousColumnsError(WrongColumnsError):
    """Exception for when a dataframe has one ore more non-specified columns."""


class MissingValuesError(ValueError):
    """Exception for when a dataframe has non-nullable columns with nulls."""


class ColumnDTypeError(TypeError):
    """Exception for when a dataframe has one or more columns with wrong dtypes."""


class RowValueError(ValueError):
    """Exception for when a dataframe has a row with a impermissible value."""


class RowDoesNotExist(RuntimeError):
    """Exception for when a single row was expected, but none were returned."""


class MultipleRowsReturned(RuntimeError):
    """Exception for when a single row was expected, but several were returned."""



"""
Module which wraps around the duckdb module in an opiniated manner.
"""
from __future__ import annotations

from collections.abc import Iterable, Iterator, Mapping
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import numpy as np
import polars as pl
import pyarrow as pa
from pydantic import create_model
from typing_extensions import Literal

from patito.pydantic import PYDANTIC_TO_DUCKDB_TYPES, Model, ModelType

try:
    import pandas as pd

    _PANDAS_AVAILABLE = True
except ImportError:
    _PANDAS_AVAILABLE = False

if TYPE_CHECKING:
    import duckdb


# Types which can be used to instantiate a DuckDB Relation object
RelationSource = Union[
    "pd.DataFrame", pl.DataFrame, Path, str, "duckdb.DuckDBPyRelation", "Relation"
]

# Used to refer to type(self) in Relation methods which preserve the type.
# Hard-coding Relation or Relation[ModelType] does not work for subclasses
# that return type(self) since that will refer to the parent class.
# See relevant SO answer: https://stackoverflow.com/a/63178532
RelationType = TypeVar("RelationType", bound="Relation")

# The SQL types supported by DuckDB
# See: https://duckdb.org/docs/sql/data_types/overview
DuckDBSQLType = Literal[
    "BIGINT",
    "BOOLEAN",
    "BLOB",
    "DATE",
    "DOUBLE",
    "DECIMAL",
    "HUGEINT",
    "INTEGER",
    "REAL",
    "SMALLINT",
    "TIME",
    "TIMESTAMP",
    "TINYINT",
    "UBIGINT",
    "UINTEGER",
    "USMALLINT",
    "UTINYINT",
    "UUID",
    "VARCHAR",
]

# Used for backward-compatible patches
try:
    POLARS_VERSION = tuple(map(int, pl.__version__.split(".")))
except ValueError:
    POLARS_VERSION = None


def create_pydantic_model(relation: "duckdb.DuckDBPyRelation") -> Type[Model]:
    """Create pydantic model deserialization of the given relation."""
    pydantic_annotations = {column: (Any, ...) for column in relation.columns}
    return create_model(
        relation.alias,
        __base__=Model,
        **pydantic_annotations,
    )


class Relation(Generic[ModelType]):
    # Can be set by subclasses in order to specify the serialization class for rows.
    # Must accept column names as keyword arguments.
    model: Optional[Type[ModelType]] = None

    # The alias that can be used to refer to the relation in queries
    alias: str
    # The columns of the relation
    columns: List[str]
    # The SQL types of the relation
    types: List[DuckDBSQLType]

    def __init__(
        self,
        relation: "duckdb.DuckDBPyRelation",
        database: Database,
        model: Optional[Type[ModelType]] = None,
    ) -> None:
        """
        Wrap around the given DuckDB Relation object associated with the given database.

        Args:
            relation (duckdb.DuckDBPyRelation): Relation to be wrapped.
            database (Database): Associated database which will be queried/mutated if
            relation is executed.
            model (Optional[Type[ModelType]]): Sub-class of patito.Model which
            specifies how to deserialize rows when fetched with methods such as .get()
            and __iter__() and how table schema should be constructed.
        """
        import duckdb

        if not isinstance(relation, duckdb.DuckDBPyRelation):
            raise TypeError(f"Invalid relation type {type(relation)}!")
        self.database = database
        self._relation = relation
        if model is not None:
            self.model = model

    def aggregate(
        self,
        *aggregations: str,
        group_by: Union[str, Iterable[str]],
        **named_aggregations: str,
    ) -> Relation:
        """
        Return relation formed by group by SQL aggregation(s).

        Args:
            aggregations: Zero or more aggregation expressions such as
                "sum(column_name)" and "count(distinct column_name)".
            group_by: A single column name or iterable collection of column names to
                group by.
            named_aggregations: Zero or more aggregated expressions where the keyword is
                used to name the given aggregation. For example,
                my_column="sum(column_name)" is inserted as
                "sum(column_name) as my_column" in the executed SQL query.
        """
        expression = ", ".join(
            aggregations
            + tuple(
                f"{expression} as {column_name}"
                for column_name, expression in named_aggregations.items()
            )
        )
        relation = self._relation.aggregate(
            aggr_expr=expression,
            group_expr=group_by if isinstance(group_by, str) else ", ".join(group_by),
        )
        return self._wrap(relation=relation, schema_change=True)

    def add_suffix(self, suffix: str) -> Relation:
        """Add suffix to all columns of relation."""
        return self.project(
            ", ".join(f"{column} as {column}{suffix}" for column in self.columns)
        )

    def add_prefix(self, prefix: str) -> Relation:
        """Add prefix to all columns of relation."""
        return self.project(
            ", ".join(f"{column} as {prefix}{column}" for column in self.columns)
        )

    def all(self, *filters: str, **equalities: Union[int, float, str]) -> bool:
        """
        Return True if the given conditions are true for all rows in the relation.

        See Relation.filter() for additional information regarding the parameters.
        """
        return self.filter(*filters, **equalities).count() == self.count()

    def case(
        self,
        *,
        from_column: str,
        to_column: str,
        mapping: Mapping[Any, Any],
        default: Union[str, int, float],
    ) -> Relation:
        """
        Map values of one column over to a new column.

        Args:
            from_column: Name of column definining the domain of the mapping.
            to_column: Name of column to insert the mapped values into.
            mapping: Dictionary defining the mapping. The dictionary keys represent the
                input values, while the dictionary values represent the output values.
                Items are inserted into the SQL case statement by their repr() string
                value.
            default: Default output value for inputs which have no provided mapping.
        """
        case_column_definition = f"case {from_column} " + (
            " ".join(f"when {key!r} then {value!r}" for key, value in mapping.items())
            + f" else {default} end as {to_column}"
        )
        new_relation = self._relation.project("*, " + case_column_definition)
        return self._wrap(relation=new_relation, schema_change=True)

    def coalesce(
        self: RelationType,
        **column_expressions: Union[str, int, float],
    ) -> RelationType:
        """
        Replace null-values in given columns with respective values.

        For example, coalesce(column_name=value) is compiled to:
        f"coalesce({column_name}, {repr(value)}) as column_name" in the resulting SQL.

        Args:
            column_expressions: Keywords indicate which columns to coalesce, while the
                string representation of the respective arguments are used as the
                null-replacement.
        """
        projections = []
        for column in self.columns:
            if column in column_expressions:
                expression = column_expressions[column]
                projections.append(f"coalesce({column}, {expression!r}) as {column}")
            else:
                projections.append(column)
        return self.project(*projections)

    def count(self) -> int:
        """Return the number of rows in the given relation."""
        return cast(Tuple[int], self._relation.aggregate("count(*)").fetchone())[0]

    def create_table(self: RelationType, name: str) -> RelationType:
        """
        Create new database table based on relation.

        If self.model is set, then the model is used to infer the table schema.
        Otherwise, a permissive table schema is created based on the relation data.
        """
        if self.model is not None:
            self.database.create_table(name=name, model=self.model)
            self.insert_into(table_name=name)
        else:
            self._relation.create(table_name=name)
        return self.database.table(name)

    def drop(self, *columns: str) -> Relation:
        """
        Drop specified column(s).

        Args:
            columns (str): Any number of string column names to be dropped.
        """
        new_columns = self.columns.copy()
        for column in columns:
            new_columns.remove(column)
        return self[new_columns]

    def execute(self) -> duckdb.DuckDBPyResult:
        """Execute built relation query and return result object."""
        # A star-select is here performed in order to work around certain DuckDB bugs
        return self._relation.project("*").execute()

    def get(self, *filters: str, **equalities: Union[str, int, float]) -> ModelType:
        """
        Fetch the single row that matches the given filter(s).

        If you expect a relation to already return one row, you can use get() without
        any arguments to return that row.

        Raises:
            RuntimeError: RuntimeError is thrown if not exactly one single row matches
            the given filter.

        Args:
            filters (str): A conjunction of SQL where clauses.
            equalities (Any): A conjunction of SQL equality clauses. The keyword name
            is the column and the parameter is the value of the equality.

        Returns:
            row (Row): A pydantic-derived base model representing the given row.
        """
        if filters or equalities:
            relation = self.filter(*filters, **equalities)
        else:
            relation = self
        result = relation.execute()
        row = cast(tuple, result.fetchone())
        if row is None or result.fetchone() is not None:
            args = ", ".join(repr(f) for f in filters)
            kwargs = ", ".join(f"{key}={value!r}" for key, value in equalities.items())
            raise RuntimeError(
                f"Relation.get({args}, {kwargs}) returned {relation.count()} rows!"
            )
        return self._to_model(row=row)

    def _to_model(self, row: tuple) -> ModelType:
        """
        Cast row tuple to proper return type.

        If self.model is set, either by a class variable of a subclass or by the
        invocation of Relation.set_model(), that type is used to construct the return
        value. Otherwise, a pydantic model is dynamically created based on the column
        schema of the relation.
        """
        kwargs = {column: value for column, value in zip(self._relation.columns, row)}
        if self.model:
            return self.model(**kwargs)
        else:
            RowModel = create_pydantic_model(relation=self._relation)
            return cast(
                ModelType,
                RowModel(**kwargs),
            )

    def filter(
        self: RelationType,
        *filters: str,
        **equalities: Union[str, int, float],
    ) -> RelationType:
        """
        Filter rows of relation.

        Args:
            filters (str): A conjunction of SQL where clauses.
            equalities (Any): A conjunction of SQL equality clauses. The keyword name
            is the column and the parameter is the value of the equality.

        Returns:
            rel (Relation): A new relation where all rows satisfy the given criteria.
        """
        clauses = []
        if filters:
            clauses.extend(filters)
        if equalities:
            clauses.extend(f"{key}={value!r}" for key, value in equalities.items())
        filter_string = " and ".join(clauses)
        return self._wrap(self._relation.filter(filter_string), schema_change=False)

    def inner_join(self: RelationType, other: RelationSource, on: str) -> RelationType:
        """
        Inner join relation with other relation source based on condition.

        Args:
            other (RelationSource): A source which can be casted to a Relation object,
            and be used as the right table in the join.
            on (str): Join condition following the "INNER JOIN ... ON" in the SQL query.

        Returns:
            relation (Relation): New relation based on the joined tables.
        """
        return self._wrap(
            self._relation.join(
                other_rel=self.database.to_relation(other)._relation,
                condition=on,
                how="inner",
            ),
            schema_change=True,
        )

    def left_join(self: RelationType, other: RelationSource, on: str) -> RelationType:
        """
        Left join relation with other relation source based on condition.

        Args:
            other (RelationSource): A source which can be casted to a Relation object,
            and be used as the right table in the join.
            on (str): Join condition following the "LEFT JOIN ... ON" in the SQL query.

        Returns:
            relation (Relation): New relation based on the joined tables.
        """
        return self._wrap(
            self._relation.join(
                other_rel=self.database.to_relation(other)._relation,
                condition=on,
                how="left",
            ),
            schema_change=True,
        )

    def insert_into(
        self: RelationType,
        table_name: str,
    ) -> RelationType:
        """
        Insert relation values into table.

        The relation must contain all the columns present in the target table.
        Extra columns are ignored and the column order is automatically matched
        with the target table.

        Args:
            table_name (str): Name of table for which to insert values into.

        Returns:
            relation: The original relation (self).
        """
        table = self.database.table(table_name)
        missing_columns = set(table.columns) - set(self.columns)
        if missing_columns:
            raise TypeError(
                f"Relation is missing column(s) {missing_columns} "
                f"in order to be inserted into table '{table_name}'!",
            )

        reordered_relation = self[table.columns]
        reordered_relation._relation.insert_into(table_name=table_name)
        return self

    def project(self, *projections: str, **named_projections: str) -> Relation:
        """
        Return relation based on one or more select projections.

        Keyword arguments are converted into "{arg} as {keyword}" SQL statements.
        For example relation.project("null as column_name") or
        relation.project("a", "b", "c is null as is_c_null", d="coalesce(a, b)").
        """
        # We expand '*' to an explicit list of columns in order to support redefining
        # columns within the star expressed columns.
        expanded_projections: list = list(projections)
        try:
            star_index = projections.index("*")
            expanded_projections[star_index : star_index + 1] = self.columns
        except ValueError:
            pass

        projection = ", ".join(
            expanded_projections
            + list(
                f"{expression} as {column_name}"
                for column_name, expression in named_projections.items()
            )
        )
        relation = self._relation.project(projection)
        return self._wrap(relation=relation, schema_change=True)

    def rename(self, **columns: str) -> Relation:
        """
        Rename columns from left-hand-side value to right-hand-side value.

        For instance, relation.rename(a="b") will rename column "a" to "b".
        """
        existing_columns = self._relation.columns
        missing = set(columns.keys()) - set(existing_columns)
        if missing:
            raise ValueError(
                f"Column '{missing.pop()}' can not be renamed as it does not exist. "
                f"The columns of the relation are: {', '.join(existing_columns)}."
            )
        # If we rename a column to overwrite another existing one, the column should
        # be overwritten.
        existing_columns = set(existing_columns) - set(columns.values())
        relation = self._relation.project(
            ", ".join(
                f"{column} as {columns.get(column, column)}"
                for column in existing_columns
            )
        )
        return self._wrap(relation=relation, schema_change=True)

    def set_model(self, model):  # noqa: ANN
        """
        Specify column schema and the constructor method for rows in the relation.

        Used by methods which return individual rows, such as Relation.get()
        and Relation.__iter__().
        """
        # We are not able to annotate the generic instance of type(self)[type(model)]
        # due to the lack of higher-kinded generics in python as of this writing.
        # See: https://github.com/python/typing/issues/548
        # This cast() will be wrong for sub-classes of Relation...
        return cast(
            Relation[model],
            type(self)(relation=self._relation, database=self.database, model=model),
        )

    @property
    def sql_types(self) -> dict[str, str]:
        """Return column name -> DuckDB SQL type dictionary mapping."""
        return dict(zip(self._relation.columns, self._relation.types))

    def to_pandas(self) -> "pd.DataFrame":
        """Return a pandas DataFrame representation of relation object."""
        return cast("pd.DataFrame", self._relation.to_df())

    def to_df(self) -> pl.DataFrame:
        """Return a polars DataFrame representation of relation object."""
        # Here we do a star-select to work around certain weird issues with DuckDB
        self._relation = self._relation.project("*")
        arrow_table = cast(pa.lib.Table, self._relation.to_arrow_table())
        if POLARS_VERSION and POLARS_VERSION <= (0, 13, 38):
            # Fix for https://github.com/pola-rs/polars/issues/3500
            schema = arrow_table.schema
            for index, field in enumerate(schema):
                if isinstance(field.type, pa.DictionaryType):
                    dict_field = field.with_type(
                        pa.dictionary(index_type=pa.int8(), value_type=pa.utf8())
                    )
                    schema = schema.set(index, dict_field)
            arrow_table = arrow_table.cast(schema)
        try:
            return cast(pl.DataFrame, pl.from_arrow(arrow_table))
        except pa.ArrowInvalid:  # pragma: no cover
            # Empty relations with enum columns can sometimes produce errors.
            # As a last-ditch effort, we convert such columns to VARCHAR.
            casted_columns = [
                f"{field.name}::VARCHAR as {field.name}"
                if isinstance(field.type, pa.DictionaryType)
                else field.name
                for field in arrow_table.schema
            ]
            non_enum_relation = self._relation.project(", ".join(casted_columns))
            arrow_table = non_enum_relation.to_arrow_table()
            return cast(pl.DataFrame, pl.from_arrow(arrow_table))

    def to_series(self) -> pl.Series:
        if len(self._relation.columns) != 1:
            raise TypeError(
                f"{self.__class__.__name__}.to_series() was invoked on a relation with "
                f"{len(self._relation.columns)} columns, while exactly 1 is required!"
            )
        dataframe = cast(pl.DataFrame, pl.from_arrow(self._relation.to_arrow_table()))
        return dataframe.to_series(index=0).alias(name=self._relation.columns[0])

    def union(self: RelationType, other: RelationSource) -> RelationType:
        """
        Produce new relation from union of two relation (sources).

        The two relations must have the same column names, but not necessarily in the
        same order as reordering of columns is automatically performed, unlike regular
        SQL.
        """
        other_relation = self.database.to_relation(other)
        if set(self.columns) != set(other_relation.columns):
            msg = "Union between relations with different column names is not allowed."
            additional_left = set(self.columns) - set(other_relation.columns)
            additional_right = set(other_relation.columns) - set(self.columns)
            if additional_left:
                msg += f" Additional columns in left relation: {additional_left}."
            if additional_right:
                msg += f" Additional columns in right relation: {additional_right}."
            raise TypeError(msg)
        reordered_relation = other_relation[self._relation.columns]
        unioned_relation = self._relation.union(reordered_relation._relation)
        return self._wrap(relation=unioned_relation, schema_change=False)

    def with_missing_defaultable_columns(self: RelationType) -> RelationType:
        """
        Add missing defaultable columns filled with the default values of correct type.

        Make sure to invoke Relation.set_model() with the correct model schema before
        executing with_missing_default_columns().
        """
        if self.model is None:
            class_name = self.__class__.__name__
            raise TypeError(
                f"{class_name}.with_missing_default_columns() invoked without "
                f"{class_name}.model having been set! "
                "You should invoke {class_name}.set_model() first!"
            )

        missing_columns = set(self.model.columns) - set(self._relation.columns)
        defaultable_columns = self.model.defaults.keys()
        missing_defaultable_columns = missing_columns & defaultable_columns

        projection = "*"
        for column_name in missing_defaultable_columns:
            sql_type = self.model.sql_types[column_name]
            default_value = self.model.defaults[column_name]
            projection += f", {default_value!r}::{sql_type} as {column_name}"

        relation = self._relation.project(projection)
        return self._wrap(relation=relation, schema_change=False)

    def with_missing_nullable_columns(self: RelationType) -> RelationType:
        """
        Add missing nullable columns filled with correctly typed nulls.

        Make sure to invoke Relation.set_model() with the correct schema before
        executing with_missing_default_columns().
        """
        if self.model is None:
            class_name = self.__class__.__name__
            raise TypeError(
                f"{class_name}.with_missing_nullable_columns() invoked without "
                f"{class_name}.model having been set! "
                "You should invoke {class_name}.set_model() first!"
            )
        missing_columns = set(self.model.columns) - set(self.columns)
        missing_nullable_columns = self.model.nullable_columns & missing_columns

        projection = "*"
        for missing_nullable_column in missing_nullable_columns:
            sql_type = self.model.sql_types[missing_nullable_column]
            projection += f", null::{sql_type} as {missing_nullable_column}"

        relation = self._relation.project(projection)
        return self._wrap(relation=relation, schema_change=False)

    def __add__(self: RelationType, other: RelationSource) -> RelationType:
        """Invoke self.union(other)."""
        return self.union(other)

    def __eq__(self, other: RelationSource) -> bool:
        """Check if Relation is equal to a Relation-able data source."""
        other_relation = self.database.to_relation(other)
        # Check if the number of rows are equal, and then check if each row is equal.
        # Use zip(self, other_relation, strict=True) when we upgrade to Python 3.10.
        return self.count() == other_relation.count() and all(
            row == other_row for row, other_row in zip(self, other_relation)
        )

    def __getattr__(self, name: str) -> Any:  # noqa: ANN
        """
        Resolve object attribute access.

        This magic method is called whenever a non-existing attribute of self is
        accessed. We delegate all attributes to the underlying DuckDB relation
        but wrap methods which return new relations.
        """
        # We do not want end-users to use these DuckDB methods, but rather:
        # create_table, to_df, and inner_join.
        if name in {"create", "df", "join"}:
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{name}'")

        schema_preserving = {
            "distinct",
            "except_",
            "intersect",
            "limit",
            "order",
            "set_alias",
        }
        schema_changing = {
            "join",
            "map",
        }
        relation_returning_methods = schema_preserving | schema_changing

        if name in relation_returning_methods:
            return lambda *args, **kwargs: self._wrap(
                relation=getattr(self._relation, name)(*args, **kwargs),
                schema_change=name in schema_changing,
            )

        try:
            return getattr(self._relation, name)
        except AttributeError:
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{name}'")

    def __getitem__(self, key: Union[str, Iterable[str]]) -> Relation:
        """
        Return Relation with selected columns.

        Uses Relation.project() under-the-hood in order to perform the selection.
        Can technically be used to rename columns, define derived columns, and
        so on, but prefer the use of Relation.project() for such use cases.

        Args:
            key (Union[str, Iterable[str]): Columns to select, either a single
            column represented as a string, or an iterable of strings.

        Returns:
            relation (Relation): The return type is not exectly Relation[ModelType],
            but rather a subset of the fields present in ModelType.
        """
        projection = key if isinstance(key, str) else ", ".join(key)
        return self._wrap(
            relation=self._relation.project(projection),
            schema_change=True,
        )

    def __iter__(self) -> Iterator[ModelType]:
        """Iterate over rows in relation as column-named tuples."""
        result = self._relation.execute()
        while True:
            row_tuple = result.fetchone()
            if not row_tuple:
                return
            else:
                yield self._to_model(row_tuple)

    def __len__(self) -> int:
        """Return number of rows in relation."""
        return self.count()

    def __str__(self) -> str:
        """
        Return string representation of Relation object.

        Includes an expression tree, the result columns, and a result preview.
        """
        return str(self._relation)

    def _wrap(
        self: RelationType,
        relation: "duckdb.DuckDBPyRelation",
        schema_change: bool = False,
    ) -> RelationType:
        """
        Wrap DuckDB Relation object in same Relation wrapper class as self.

        This will preserve the type of the relation, even for subclasses Relation.
        It should therefore only be used for relations which can be considered schema-
        compatible with the original relation. Otherwise set schema_change to True
        in order to create a Relation base object instead.
        """
        return type(self)(
            relation=relation,
            database=self.database,
            model=self.model if not schema_change else None,
        )


class Database:
    # Method for executing SQL queries which do not return relations
    execute: Callable[[str], None]

    # Method for executing arbitrary select queries which return relations
    query: Callable[[str], Relation]

    # Method for extracting a relation referring to a table
    table: Callable[[str], Relation]

    def __init__(self, path: Optional[Path] = None) -> None:
        """
        Instantiate a new in-memory DuckDB database.

        Args:
            path: Optional path to store all the data to. If None, the data is persisted
                in-memory only.
        """
        import duckdb

        if path:
            self.connection = duckdb.connect(database=str(path))
            self.path = path
        else:
            self.connection = duckdb.connect(database=":memory:")

    def to_relation(
        self,
        derived_from: RelationSource,
    ) -> Relation:
        """
        Create a new relation object based on data source.

        Args:
            derived_from (RelationSource): One of either a pandas DataFrame,
            a pathlib.Path to a parquet or CSV file, a SQL query string,
            or an existing relation.
        """
        import duckdb

        if isinstance(derived_from, Relation):
            if derived_from.database is not self:
                raise ValueError(
                    "Relations can't be casted between database connections."
                )
            return derived_from
        elif isinstance(derived_from, duckdb.DuckDBPyRelation):
            relation = derived_from
        elif isinstance(derived_from, str):
            relation = self.connection.from_query(derived_from)
        elif _PANDAS_AVAILABLE and isinstance(derived_from, pd.DataFrame):
            # We must replace pd.NA with np.nan in order for it to be considered
            # as null by DuckDB. Otherwise it will casted to the string <NA>
            # or even segfault.
            derived_from = cast(pd.DataFrame, derived_from).fillna(np.nan)
            relation = self.connection.from_df(derived_from)
        elif isinstance(derived_from, pl.DataFrame):
            relation = self.connection.from_arrow(derived_from.to_arrow())
        elif isinstance(derived_from, Path):
            if derived_from.suffix == ".parquet":
                relation = self.connection.from_parquet(str(derived_from))
            elif derived_from.suffix == ".csv":
                relation = self.connection.from_csv_auto(str(derived_from))
            else:
                raise ValueError(
                    f"Unsupported file suffix {derived_from.suffix} for data import!"
                )
        else:
            raise TypeError
        return Relation(relation, database=self)

    def empty_relation(self, schema: Type[ModelType]) -> Relation[ModelType]:
        """
        Create relation with zero rows, but correct schema that matches the model.

        Args:
            schema: A patito model which specifies the column names and types of the
                given relation.
        """
        return self.to_relation(schema.examples()).limit(0)

    def create_table(
        self,
        name: str,
        model: Type[ModelType],
    ) -> Relation[ModelType]:
        """
        Create table with schema matching the provided model.

        Args:
            name: Name of new table.
            model: Pydantic-derived model indicating names and types of table columns.
        Returns:
            relation (Relation[ModelType]): Relation pointing to the new table.
        """
        schema = model.schema()
        non_nullable = schema["required"]
        columns = []
        for column_name, props in model.schema()["properties"].items():
            if "enum" in props and props["type"] == "string":
                enum_values = ", ".join(repr(value) for value in props["enum"])
                enum_type_name = f"{schema['title']}__{column_name}".lower()
                self.connection.execute(
                    f"create type {enum_type_name} as enum ({enum_values})"
                )
                column = f"{column_name} {enum_type_name}"
                if column_name in non_nullable:
                    column += " not null"
            else:
                sql_type = PYDANTIC_TO_DUCKDB_TYPES[props["type"]]
                column = f"{column_name} {sql_type}"
                if column_name in non_nullable:
                    column += " not null"
            columns.append(column)
        self.connection.execute(f"create table {name} ({','.join(columns)})")
        # TODO: Fix typing
        return self.table(name).set_model(model)  # type: ignore

    def create_view(
        self,
        name: str,
        data: RelationSource,
    ) -> Relation:
        """Create a view based on the given data source."""
        return self.to_relation(derived_from=data).create_view(name)

    def __getattr__(
        self, name: str
    ) -> Union[duckdb.DuckDBPyRelation, Relation, Callable[..., Relation]]:
        """
        Resolve object attribute access.

        This magic method is called whenever a non-existing attribute of self is
        accessed. We delegate all attributes to the underlying DuckDB connection,
        but wrap methods which return relations in the Relations wrapper.
        """
        relation_methods = {
            "from_arrow_table",
            "from_csv_auto",
            "from_df",
            "from_parquet",
            "from_query",
            "query",
            "table",
            "table_function",
            "values",
            "view",
        }
        if name in relation_methods:
            return lambda *args, **kwargs: Relation(
                getattr(self.connection, name)(*args, **kwargs), database=self
            )
        else:
            return getattr(self.connection, name)

    def __contains__(self, table: str) -> bool:
        """Return True if the database contains a table with the given name."""
        try:
            self.connection.table(table_name=table)
            return True
        except RuntimeError:
            return False



"""Module for validating datastructures with respect to model specifications."""
from __future__ import annotations

from typing import TYPE_CHECKING, Type, Union, cast

import polars as pl

from patito.exceptions import (
    ColumnDTypeError,
    ErrorWrapper,
    MissingColumnsError,
    MissingValuesError,
    RowValueError,
    SuperflousColumnsError,
    ValidationError,
)

try:
    import pandas as pd

    _PANDAS_AVAILABLE = True
except ImportError:
    _PANDAS_AVAILABLE = False

if TYPE_CHECKING:
    from patito import Model


VALID_POLARS_TYPES = {
    "enum": {pl.Categorical},
    "boolean": {pl.Boolean},
    "string": {pl.Utf8, pl.Datetime, pl.Date},
    "number": {pl.Float32, pl.Float64},
    "integer": {
        pl.Int8,
        pl.Int16,
        pl.Int32,
        pl.Int64,
        pl.UInt8,
        pl.UInt16,
        pl.UInt32,
        pl.UInt64,
    },
}


def _find_errors(  # noqa: C901
    dataframe: pl.DataFrame,
    schema: Type[Model],
) -> list[ErrorWrapper]:
    """
    Validate the given dataframe.

    Args:
        dataframe: Polars DataFrame to be validated.
        schema: Patito model which specifies how the dataframe should be structured.

    Returns:
        A list of patito.exception.ErrorWrapper instances. The specific validation
        error can be retrieved from the "exc" attribute on each error wrapper instance.

        MissingColumnsError: If there are any missing columns.
        SuperflousColumnsError: If there are additional, non-specified columns.
        MissingValuesError: If there are nulls in a non-optional column.
        ColumnDTypeError: If any column has the wrong dtype.
        NotImplementedError: If validation has not been implement for the given
            type.
    """
    errors: list[ErrorWrapper] = []
    # Check if any columns are missing
    for missig_column in set(schema.columns) - set(dataframe.columns):
        errors.append(
            ErrorWrapper(
                MissingColumnsError("Missing column"),
                loc=missig_column,
            )
        )

    # Check if any additional columns are included
    for superflous_column in set(dataframe.columns) - set(schema.columns):
        errors.append(
            ErrorWrapper(
                SuperflousColumnsError("Superflous column"),
                loc=superflous_column,
            )
        )

    # Check if any non-optional columns have null values
    for column in schema.non_nullable_columns.intersection(dataframe.columns):
        num_missing_values = dataframe.get_column(name=column).null_count()
        if num_missing_values:
            errors.append(
                ErrorWrapper(
                    MissingValuesError(
                        f"{num_missing_values} missing "
                        f"{'value' if num_missing_values == 1 else 'values'}"
                    ),
                    loc=column,
                )
            )

    # Check if any column has a wrong dtype
    valid_dtypes = schema.valid_dtypes
    model_schema = schema.schema()
    dataframe_datatypes = dict(zip(dataframe.columns, dataframe.dtypes))
    for column_name, column_properties in model_schema["properties"].items():
        if column_name not in dataframe.columns:
            continue

        polars_type = dataframe_datatypes[column_name]
        if polars_type not in valid_dtypes[column_name]:
            errors.append(
                ErrorWrapper(
                    ColumnDTypeError(
                        f"Polars dtype {polars_type} does not match model field type."
                    ),
                    loc=column_name,
                )
            )

        # Test for when only specific values are accepted
        if "enum" in column_properties:
            permissible_values = set(column_properties["enum"])
            actual_values = set(dataframe[column_name].unique())
            impermissible_values = actual_values - permissible_values
            if impermissible_values:
                errors.append(
                    ErrorWrapper(
                        RowValueError(
                            f"Rows with invalid values: {impermissible_values}."
                        ),
                        loc=column_name,
                    )
                )

        if column_properties.get("unique", False):
            num_duplicated = dataframe[column_name].is_duplicated().sum()
            if num_duplicated > 0:
                errors.append(
                    ErrorWrapper(
                        RowValueError(f"{num_duplicated} rows with duplicated values."),
                        loc=column_name,
                    )
                )

        # Check for bounded value fields
        col = pl.col(column_name)
        filters = {
            "maximum": lambda v: col <= v,
            "exclusiveMaximum": lambda v: col < v,
            "minimum": lambda v: col >= v,
            "exclusiveMinimum": lambda v: col > v,
            "multipleOf": lambda v: (col == 0) | ((col % v) == 0),
            "const": lambda v: col == v,
            "pattern": lambda v: col.str.contains(v),
            "minLength": lambda v: col.str.lengths() >= v,
            "maxLength": lambda v: col.str.lengths() <= v,
        }
        checks = [
            check(column_properties[key])
            for key, check in filters.items()
            if key in column_properties
        ]
        if checks:
            lazy_df = dataframe.lazy()
            for check in checks:
                lazy_df = lazy_df.filter(check)
            valid_rows = lazy_df.collect()
            invalid_rows = dataframe.height - valid_rows.height
            if invalid_rows > 0:
                errors.append(
                    ErrorWrapper(
                        RowValueError(
                            f"{invalid_rows} row{'' if invalid_rows == 1 else 's'} "
                            "with out of bound values."
                        ),
                        loc=column_name,
                    )
                )

        if "constraints" in column_properties:
            custom_constraints = column_properties["constraints"]
            if isinstance(custom_constraints, pl.Expr):
                custom_constraints = [custom_constraints]
            illegal_rows = dataframe.filter(
                pl.all([constraint.is_not() for constraint in custom_constraints])
            )
            if illegal_rows.height > 0:
                errors.append(
                    ErrorWrapper(
                        RowValueError(
                            f"{illegal_rows.height} "
                            f"row{'' if illegal_rows.height == 1 else 's'} "
                            "does not match custom constraints."
                        ),
                        loc=column_name,
                    )
                )

    return errors


def validate(
    dataframe: Union["pd.DataFrame", pl.DataFrame], schema: Type[Model]
) -> None:
    """
    Validate the given dataframe.

    Args:
        dataframe: Polars DataFrame to be validated.
        schema: Patito model which specifies how the dataframe should be structured.

    Raises:
        ValidationError: If the given dataframe does not match the given schema.
    """
    if _PANDAS_AVAILABLE and isinstance(dataframe, pd.DataFrame):
        polars_dataframe = pl.from_pandas(dataframe)
    else:
        polars_dataframe = cast(pl.DataFrame, dataframe)

    errors = _find_errors(dataframe=polars_dataframe, schema=schema)
    if errors:
        raise ValidationError(errors=errors, model=schema)
